#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional

from ..api.image import ImageLabel
from ..common.logging import get_logger
from ..common.utils import to_semantic_version
from ..estimator import Estimator
from ..session import Session

logger = get_logger(__name__)


class HuggingFaceEstimator(Estimator):
    """Handle training of custom HuggingFace model.

    The HuggingFace Estimator is optimized to run a HuggingFace training
    script in the PAI Training Service with a specific image.

    Example::

        est = HuggingFaceEstimator(
            source_dir="./train/src/",
            command="python train.py",
            transformers_version = 'latest',
            instance_type="ecs.c6.xlarge",
        )

        est.fit()

        print(est.model_data())

    """

    def __init__(
        self,
        command: str,
        source_dir: Optional[str] = None,
        git_config: Optional[Dict[str, str]] = None,
        image_uri: Optional[str] = None,
        transformers_version: Optional[str] = None,
        hyperparameters: Optional[Dict[str, Any]] = None,
        base_job_name: Optional[str] = None,
        checkpoints_path: Optional[str] = None,
        output_path: Optional[str] = None,
        instance_type: Optional[str] = None,
        instance_count: int = 1,
        session: Optional[Session] = None,
        **kwargs,
    ):
        """Initialize a HuggingFace Estimator constructor.

        Args:
            command (str): The command used to run the training job.
            source_dir (str, optional): The local source code directory used in the
                training job. The directory will be packaged and uploaded to an OSS
                bucket, then downloaded to the `/ml/usercode` directory in the training
                job container. If there is a `requirements.txt` file in the source code
                directory, the corresponding dependencies will be installed before the
                training script runs.

                If 'git_config' is provided, 'source_dir' should be a relative location
                to a directory in the Git repo. With the following GitHub repo directory
                structure:

                .. code::

                    |----- README.md
                    |----- src
                             |----- train.py
                             |----- test.py

                if you need 'src' directory as the source code directory, you can assign
                source_dir='./src/'.
            git_config (Dict[str, str]): Git configuration used to clone the repo.
                Including ``repo``, ``branch``, ``commit``, ``username``, ``password``
                and ``token``. The ``repo`` is required. All other fields are optional.
                ``repo`` specifies the Git repository. If you don't provide ``branch``,
                the default value 'master' is used. If you don't provide ``commit``, the
                latest commit in the specified branch is used. ``username``, ``password``
                and ``token`` are for authentication purpose.
                For example, the following config:

                .. code:: python

                    git_config = {
                        'repo': 'https://github.com/huggingface/transformers.git',
                        'branch': 'main',
                        'commit': '5ba0c332b6bef130ab6dcb734230849c903839f7'
                    }

                results in cloning the git repo specified in 'repo', then checking out
                the 'main' branch, and checking out the specified commit.
            image_uri (str, optional): If specified, the estimator will use this image
                in the training job, instead of selecting the appropriate PAI official
                image based on transformers_version. It can be an image provided by PAI
                or a user customized image. To view the images provided by PAI, please
                refer to the document:
                https://help.aliyun.com/document_detail/202834.htm.

                If ``transformers_version`` is ``None``, then ``image_uri`` is required.
                If also ``None``, then a ``ValueError`` will be raised.
            transformers_version (str, optional): Transformers version you want to use for
                executing your model training code. Defaults to ``None``. Required unless
                ``image_uri`` is provided.
            hyperparameters (dict, optional): A dictionary that represents the
                hyperparameters used in the training job. The hyperparameters will be
                stored in the `/ml/input/config/hyperparameters.json` as a JSON
                dictionary in the training container.
            base_job_name (str, optional): The base name used to generate the training
                job name.
            checkpoints_path (str, optional): An OSS URI that stores the checkpoint of the
                training job. If provided, the OSS URI will be mounted to the directory
                `/ml/output/checkpoints/`.
            output_path (str, optional): An OSS URI to store the outputs of the training
                jobs. If not provided, an OSS URI will be generated using the default
                OSS bucket in the session. When the `estimator.fit` method is called,
                a specific OSS URI under the output_path for each channel is generated
                and mounted to the training container.

                A completed training container directory structure example::

                    /ml
                    |-- usercode            			// User source code directory.
                    |   |-- requirements.txt
                    |   `-- train.py
                    |-- input               			// TrainingJob input
                    |   `-- config
                    |       |-- hyperparameters.json	// Hyperparameters in JSON
                    |       |                           // dictionary format for the
                    |       |                           // TrainingJob
                    |       |
                    |   `-- data            			// TrainingJob input channels
                    |       |                           // `/ml/input/data/` is a input
                    |       |                           // channel, and the directory
                    |       |                           // name is the channel name.
                    |       |                           // Each directory under the
                    |       |-- test-data
                    |       |   `-- test.csv
                    |       `-- train-data
                    |           `-- train.csv
                    `-- output              			// TrainingJob output channels.
                            |                           // Each directory under the
                            |                           // `/ml/output/` is an output
                            |                           // channel, and the directory
                            |                           // name is the channel name.
                            `-- model
                            `-- checkpoints
            instance_type (str): The machine instance type used to run the training job.
                To view the supported machine instance types, please refer to the
                document:
                https://help.aliyun.com/document_detail/171758.htm#section-55y-4tq-84y.
                If the instance_type is "local", the training job is executed locally
                using docker.
            instance_count (int): The number of machines used to run the training job.
            session (:class:`pai.session.Session`, optional): A pai session object manages
                interactions with PAI REST API.

            **kwargs: Additional kwargs passed to the :class:`~pai.estimator.Estimator`
                constructor.

        .. tip::

            You can find additional parameters for initializing this class at
            :class:`~pai.estimator.Estimator`.
        """
        self._validate_image_uri(
            image_uri=image_uri, transformers_version=transformers_version
        )
        self.image_uri = image_uri
        self.transformers_version = transformers_version

        super(HuggingFaceEstimator, self).__init__(
            image_uri=self.image_uri,
            command=command,
            source_dir=source_dir,
            git_config=git_config,
            hyperparameters=hyperparameters,
            base_job_name=base_job_name,
            checkpoints_path=checkpoints_path,
            output_path=output_path,
            instance_type=instance_type,
            instance_count=instance_count,
            session=session,
            **kwargs,
        )
        # Check image_uri and transformers_version
        self.training_image_uri()

    def _validate_image_uri(self, image_uri: str, transformers_version: str) -> None:
        """Check if image_uri or transformers_version arguments are specified."""
        if not image_uri and not transformers_version:
            raise ValueError(
                "transformers_version, and image_uri are both None. "
                "Specify either transformers_version or image_uri."
            )

    def training_image_uri(self) -> str:
        """Return the Docker image to use for training.

        The :meth:`~pai.estimator.Estimator.fit` method, which does the model training,
        calls this method to find the image to use for model training.

        Returns:
            str: The URI of the Docker image.
        """
        if self.image_uri:
            return self.image_uri

        labels = [
            ImageLabel.OFFICIAL_LABEL,
            ImageLabel.DLC_LABEL,
            ImageLabel.PROVIDER_COMMUNITY_LABEL,
            ImageLabel.DEVICE_TYPE_GPU,
            ImageLabel.framework_version("PyTorch", "*"),
        ]

        # Filter images by Transformers version
        if self.transformers_version == "latest":
            latest_version = self._get_latest_tf_version_for_training()
            labels.append(ImageLabel.framework_version("Transformers", latest_version))
        else:
            labels.append(
                ImageLabel.framework_version("Transformers", self.transformers_version)
            )

        resp = self.session.image_api.list(
            labels=labels,
            workspace_id=0,
            verbose=True,
        )

        if resp.total_count == 0:
            raise ValueError(
                "No official image found for Transformers version:"
                f" {self.transformers_version}. Currently supported versions are:"
                f" {self._get_supported_tf_versions_for_training()}"
            )

        image = resp.items[0]["ImageUri"]
        return image

    def _get_supported_tf_versions_for_training(self) -> List[str]:
        """Return the list of supported Transformers versions for training."""
        label_keys = "system.framework.Transformers"
        label_filter = [
            ImageLabel.OFFICIAL_LABEL,
            ImageLabel.DLC_LABEL,
            ImageLabel.PROVIDER_COMMUNITY_LABEL,
            ImageLabel.DEVICE_TYPE_GPU,
            ImageLabel.framework_version("PyTorch", "*"),
            ImageLabel.framework_version("Transformers", "*"),
        ]
        list_image_labels = self.session.image_api.list_labels(
            label_keys=label_keys,
            label_filter=label_filter,
            workspace_id=0,
        )

        res = []
        for label in list_image_labels:
            if label["Value"] not in res:
                res.append(label["Value"])

        res.sort(key=lambda x: to_semantic_version(x))
        return res

    def _get_latest_tf_version_for_training(self) -> str:
        """Return the latest Transformers version for training."""
        res = self._get_supported_tf_versions_for_training()
        return max(
            res,
            key=lambda x: to_semantic_version(x),
        )



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

from .artifact import (
    ArtifactMetadataUtils,
    DataType,
    LocationArtifactMetadata,
    LocationType,
    ModelType,
    PipelineArtifact,
)
from .parameter import ParameterType, PipelineParameter
from .spec import IO_TYPE_INPUTS, IO_TYPE_OUTPUTS, InputsSpec, OutputsSpec
from .variable import PipelineVariable

__all__ = [
    "PipelineVariable",
    "PipelineArtifact",
    "ModelType",
    "DataType",
    "LocationType",
    "LocationArtifactMetadata",
    "PipelineParameter",
    "ParameterType",
    "InputsSpec",
    "OutputsSpec",
    "ArtifactMetadataUtils",
    "IO_TYPE_OUTPUTS",
    "IO_TYPE_INPUTS",
]



#  Copyright 2024 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from .common.consts import JobType, StoragePathCategory
from .common.logging import get_logger
from .common.utils import experimental, random_str, to_plain_text
from .job import (
    AlgorithmSpec,
    Channel,
    CodeDir,
    ExperimentConfig,
    SpotSpec,
    TrainingJob,
    UriOutput,
    UserVpcConfig,
    _TrainingJobSubmitter,
)
from .job._training_job import ResourceType
from .session import Session, get_default_session

logger = get_logger(__name__)


@experimental
class Processor(_TrainingJobSubmitter):
    def __init__(
        self,
        image_uri: str,
        command: Union[str, List[str]],
        source_dir: Optional[str] = None,
        job_type: str = JobType.PyTorchJob,
        parameters: Optional[Dict[str, Any]] = None,
        environments: Optional[Dict[str, str]] = None,
        requirements: Optional[List[str]] = None,
        max_run_time: Optional[int] = None,
        base_job_name: Optional[str] = None,
        output_path: Optional[str] = None,
        instance_type: Optional[str] = None,
        spot_spec: Optional[SpotSpec] = None,
        resource_type: Optional[Union[str, ResourceType]] = None,
        instance_count: Optional[int] = None,
        user_vpc_config: Optional[UserVpcConfig] = None,
        experiment_config: Optional[ExperimentConfig] = None,
        settings: Optional[Dict[str, Any]] = None,
        labels: Optional[Dict[str, str]] = None,
        session: Optional[Session] = None,
    ):
        """Processor constructor.

        Args:
            image_uri (str): The image used in the job. It can be an image
                provided by PAI or a user customized image. To view the images provided
                by PAI, please refer to the document:
                https://help.aliyun.com/document_detail/202834.htm.
            command (Union[str, List[str]): The command used to run the job.
            source_dir (str, optional): The local source code directory used in the
                job. The directory will be packaged and uploaded to an OSS
                bucket, then downloaded to the `/ml/usercode` directory in the
                job container. If there is a `requirements.txt` file in the source code
                directory, the corresponding dependencies will be installed before the
                script runs.

                If 'git_config' is provided, 'source_dir' should be a relative location
                to a directory in the Git repo. With the following GitHub repo directory
                structure:

                .. code::

                    |----- README.md
                    |----- src
                             |----- train.py
                             |----- test.py

                if you need 'src' directory as the source code directory, you can assign
                source_dir='./src/'.
            job_type (str): The type of job, which can be TFJob, PyTorchJob, XGBoostJob,
                etc. Default value is PyTorchJob.
            parameters (dict, optional): A dictionary that represents the
                parameters used in the job. The parameters will be
                stored in the `/ml/input/config/hyperparameters.json` as a JSON
                dictionary in the container.
            environments: A dictionary that maps environment variable names to their values.
                This optional field allows you to provide a set of environment variables that will be
                applied to the context where the code is executed.
            requirements (list, optional): An optional list of strings that specifies the Python
                package dependencies with their versions. Each string in the list should be in the format
                'package' or 'package==version'. This is similar to the contents of a requirements.txt file used
                in Python projects. If requirements.txt is provided in user code directory, requirements
                will override the conflict dependencies directly.
            max_run_time (int, optional): The maximum time in seconds that the
                job can run. The job will be terminated after the time is
                reached (Default None).
            base_job_name (str, optional): The base name used to generate the
                job name.
            output_path (str, optional): An OSS URI to store the outputs of the
                jobs. If not provided, an OSS URI will be generated using the default
                OSS bucket in the session. When the `estimator.fit` method is called,
                a specific OSS URI under the output_path for each channel is generated
                and mounted to the container.

                A completed container directory structure example::

                    /ml
                    |-- usercode            			// User source code directory.
                    |   |-- requirements.txt
                    |   `-- train.py
                    |-- input               			// Job input
                    |   `-- config
                    |       |-- hyperparameters.json	// Hyperparameters in JSON
                    |       |                           // dictionary format for the
                    |       |                           // Job
                    |       |
                    |   `-- data            			// Job input channels
                    |       |                           // `/ml/input/data/` is a input
                    |       |                           // channel, and the directory
                    |       |                           // name is the channel name.
                    |       |                           // Each directory under the
                    |       |-- test-data
                    |       |   `-- test.csv
                    |       `-- train-data
                    |           `-- train.csv
                    `-- output              			// Job output channels.
                            |                           // Each directory under the
                            |                           // `/ml/output/` is an output
                            |                           // channel, and the directory
                            |                           // name is the channel name.
                            `-- model
                            `-- checkpoints

            instance_type (str): The machine instance type used to run the job.
                To view the supported machine instance types, please refer to the
                document:
                https://help.aliyun.com/document_detail/171758.htm#section-55y-4tq-84y.
                If the instance_type is "local", the job is executed locally
                using docker.
            instance_count (int): The number of machines used to run the job.
            resource_type (str, optional): The resource type used to run the training job.
                By default, general computing resource is used. If the resource_type is
                'Lingjun', Lingjun computing resource is used.
            user_vpc_config (:class:`pai.estimator.UserVpcConfig`, optional): The VPC
                configuration used to enable the job instance to connect to the
                specified user VPC. If provided, an Elastic Network Interface (ENI) will
                be created and attached to the job instance, allowing the
                instance to access the resources within the specified VPC. Default to
                None.
            experiment_config(:class:`pai.estimator.ExperimentConfig`, optional): The
                experiment configuration used to construct the relationship between the
                job and the experiment. If provided, the training job will belong to the
                specified experiment, in which case the job will use artifact_uri of
            settings (dict, optional): A dictionary that represents the additional settings
                for job, such as AIMaster configurations.
                experiment as default output path. Default to None.
            labels (Dict[str, str], optional): A dictionary that maps label names to
                their values. This optional field allows you to provide a set of labels
                that will be applied to the training job.
            session (Session, optional): A PAI session instance used for communicating
                with PAI service.

        """
        self.image_uri = image_uri
        self.command = command
        self.source_dir = source_dir
        self.job_type = job_type or JobType.PyTorchJob
        self.parameters = parameters or dict()
        self.session = session or get_default_session()

        self._input_channels = None
        self._output_channels = None
        super().__init__(
            resource_type=resource_type,
            spot_spec=spot_spec,
            base_job_name=base_job_name,
            output_path=output_path,
            experiment_config=experiment_config,
            instance_type=instance_type,
            instance_count=instance_count or 1,
            user_vpc_config=user_vpc_config,
            max_run_time=max_run_time,
            environments=environments,
            requirements=requirements,
            labels=labels,
            settings=settings,
        )

    def run(
        self,
        inputs: Dict[str, Any] = None,
        outputs: Dict[str, Any] = None,
        wait: bool = True,
        show_logs: bool = True,
    ) -> TrainingJob:
        """Submit a job with the given input and output channels.

        Args:
            inputs (Dict[str, Any]): A dictionary representing the input data for the
                job. Each key/value pair in the dictionary is an input channel,
                the key is the channel name, and the value is the input data. The input
                data can be an OSS URI or a NAS URI object and will be mounted to the
                `/ml/input/data/{channel_name}` directory in the job container.
            outputs (Dict[str, Any]): A dictionary representing the output data for the
                job. Each key/value pair in the dictionary is an output channel,
                the key is the channel name, and the value is the output path. The output
                path can be an OSS URI or a NAS URI object and will be mounted to the
                `/ml/outputs/data/{channel_name}` directory in the job container.
            wait (bool): Specifies whether to block until the training job is completed,
                either succeeded, failed, or stopped. (Default True).
            show_logs (bool): Whether to show the logs of the job. Default to True.
                Note that the logs will be shown only when the `wait` is set to True.

        Returns:
            :class:`pai.job.TrainingJob`: A submitted training job.

        Raises:
            UnExpectedStatusException: If the job fails.

        """
        inputs = inputs or dict()
        outputs = outputs or dict()
        job_name = self._gen_job_display_name()

        code_dest = Session.get_storage_path_by_category(
            StoragePathCategory.ProcessingSrc, to_plain_text(job_name)
        )
        code_dir = self._build_code_input(job_name, self.source_dir, code_dest)
        algo_spec = self._build_algorithm_spec(
            code_input=code_dir,
        )
        inputs = self.build_inputs(inputs, input_channels=algo_spec.input_channels)
        outputs = self.build_outputs(
            job_name=job_name,
            output_channels=algo_spec.output_channels,
            outputs=outputs,
        )

        return self._submit(
            instance_count=self.instance_count,
            instance_type=self.instance_type,
            job_name=job_name,
            hyperparameters=self.parameters,
            environments=self.environments,
            requirements=self.requirements,
            max_run_time=self.max_run_time,
            inputs=inputs,
            outputs=outputs,
            algorithm_spec=algo_spec,
            user_vpc_config=(
                self.user_vpc_config.model_dump() if self.user_vpc_config else None
            ),
            experiment_config=(
                self.experiment_config.model_dump() if self.experiment_config else None
            ),
            labels=self.labels,
            wait=wait,
            show_logs=show_logs,
        )

    def _gen_job_display_name(self, job_name=None):
        """Generate job display name."""
        if job_name:
            return job_name
        ts = datetime.now().strftime("%Y%m%d_%H%M%S")
        return "{}_{}".format(self.base_job_name or "processing_job", ts)

    def _build_algorithm_spec(self, code_input: CodeDir) -> AlgorithmSpec:
        """Build a temporary AlgorithmSpec used for submitting the Job."""

        algorithm_spec = AlgorithmSpec(
            command=(
                self.command
                if isinstance(self.command, list)
                else [
                    "sh",
                    "-c",
                    self.command,
                ]
            ),
            image=self.image_uri,
            job_type=self.job_type,
            code_dir=code_input,
            input_channels=self._input_channels or [],
            output_channels=self._output_channels or [],
        )
        return algorithm_spec

    def _training_job_base_output(self, job_name: str) -> str:
        """Generate the base output path for the job."""

        bucket_name = self.session.oss_bucket.bucket_name
        # replace non-alphanumeric character in job name.
        job_name = to_plain_text(job_name)

        if self.output_path:
            return os.path.join(self.output_path, f"{job_name}_{random_str(6)}")

        job_output_path = Session.get_storage_path_by_category(
            StoragePathCategory.ProcessingJob, f"{job_name}_{random_str(6)}"
        )
        return f"oss://{bucket_name}/{job_output_path}"

    def get_outputs_data(self) -> Dict[str, str]:
        """Show all outputs data paths.

        Returns:
            dict[str, str]: A dictionary of all outputs data paths.
        """
        if not self.latest_job:
            raise RuntimeError("Current no Job for the processor.")

        return {
            ch.name: ch.output_uri if isinstance(ch, UriOutput) else ch.dataset_id
            for ch in self.latest_job.outputs
        }

    def set_input_channels(self, channels: List[Channel]):
        self._input_channels = channels

    def set_output_channels(self, channels: List[Channel]):
        self._output_channels = channels



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import absolute_import

import itertools

import six

from pai.pipeline.types import PipelineVariable
from pai.pipeline.types.parameter import ConditionExpr, LoopItems, PipelineParameter
from pai.pipeline.types.spec import load_input_output_spec


class _ComponentRefer(object):
    def __init__(self):
        pass

    def to_dict(self):
        pass


class RegisteredComponentRefer(_ComponentRefer):
    def __init__(self, identifier, provider, version):
        super(RegisteredComponentRefer, self).__init__()
        self.identifier = identifier
        self.provider = provider
        self.version = version

    def to_dict(self):
        return {
            "identifier": self.identifier,
            "provider": self.provider,
            "version": self.version,
        }


class UnRegisteredComponentRefer(_ComponentRefer):
    def __init__(self, guid):
        super(UnRegisteredComponentRefer, self).__init__()
        self.guid = guid

    def to_dict(self):
        return {"guid": self.guid}


class PipelineStep(object):
    """Represents an execution step in PAI pipeline.

    Pipeline steps can be configured together to construct a Pipeline, which is present as workflow
    in PAI ML pipeline service.

    """

    def __init__(
        self,
        inputs=None,
        name=None,
        depends=None,
        component=None,
    ):
        """Construct a step which represent component execution in pipeline.

        Args:
            inputs (dict): Inputs for the step in dict: key is the component input name, value
                could be the output artifact/parameter from other step, input of the pipeline,
                or actual value for the step.
            name (str): Name of the step in pipeline, must be unique in the pipeline.
            depends (list): A list of PipelineStep which step depends.
            component (OperatorBase): The component used by the constructed step.
        """
        from pai.pipeline.component import RegisteredComponent

        self._depends = depends or set()
        if any([type(x) for x in self._depends if not isinstance(x, PipelineStep)]):
            raise ValueError("Invalid variable in depends, expected PipelineStep")

        self._assigned = set()

        if isinstance(component, RegisteredComponent):
            self.component_ref = RegisteredComponentRefer(
                identifier=component.identifier,
                version=component.version,
                provider=component.provider,
            )
        else:
            self.component_ref = UnRegisteredComponentRefer(guid=component.guid)

        self._name = name
        self._component = component

        (
            inputs_spec,
            outputs_spec,
        ) = load_input_output_spec(self, component.io_spec_to_dict())
        self.parent = None
        self.inputs = inputs_spec
        self.outputs = outputs_spec

        self._assign_inputs(inputs)
        self._repeated_artifact_config = {}

    @property
    def is_component_registered(self):
        if isinstance(self.component_ref, RegisteredComponentRefer):
            return True
        return False

    @property
    def component(self):
        return self._component

    def gen_name_prefix(self):
        if self.is_component_registered:
            return self.component_ref.identifier
        return self.component.name

    @classmethod
    def from_registered_component(
        cls,
        identifier,
        provider=None,
        version="v1",
        inputs=None,
        name=None,
        depends=None,
    ):
        """Build the PipelineStep from the given registered component reference: identifier, version, provider.

        Args:
            identifier: Identifier of the registered component.
            provider: Provider of the registered component.
            version: Version of the registered component.
            inputs: Inputs for the building step.
            name: Name for the building step.
            depends: Depended steps of the building step.

        Returns:
            PipelineStep: The built step instantiates from the given registered component and inputs.
        """
        from pai.pipeline.component import RegisteredComponent
        from pai.session import get_default_session

        provider = provider or get_default_session().provider
        c = RegisteredComponent.get_by_identifier(
            identifier=identifier, provider=provider, version=version
        )

        if not c:
            raise ValueError(
                "Specific register component not found: identifier={0}, provider={1}, version={2}".format(
                    identifier, provider, version
                )
            )
        return cls(
            inputs=inputs or [],
            name=name,
            depends=depends,
            component=c,
        )

    @property
    def repeated_io_config(self):
        return self._repeated_artifact_config

    def set_artifact_count(self, artifact_name, count):
        artifacts = {
            item.name: item
            for item in itertools.chain(self.outputs.artifacts, self.inputs.artifacts)
        }
        artifact = artifacts.get(artifact_name)
        if not artifact:
            raise ValueError("artifact is not exists: %s" % artifact_name)

        if not artifact.repeated:
            raise ValueError("artifact is not repeated: %s", artifact_name)
        artifact.count = count
        return self

    # TODO: Confirm pipeline step name restriction
    @classmethod
    def _validate_name(cls, name):
        if name is None:
            return
        if not isinstance(name, six.string_types):
            raise ValueError("PipelineStep name should be string type")
        if not name:
            raise ValueError("PipelineStep name should not be empty str")
        if len(name) > 30:
            raise ValueError("Given invalid pipeline step name.")

    def _assign_inputs(self, inputs):
        """Assign inputs to the step.

        Inputs could be inputs definition of pipeline, output of other steps or actual value.

        Args:
            inputs: Inputs for the step.

        """
        if not inputs:
            return
        assign_items = self.inputs.assign(inputs)
        self._assigned = self._assigned.union(set(item.name for item in assign_items))

        if isinstance(inputs, dict):
            inputs = inputs.values()

        values = []
        for ipt in inputs:
            if isinstance(ipt, (list, tuple)):
                values.extend(ipt)
            else:
                values.append(ipt)

        def _depend_step(input):
            from pai.pipeline.types.artifact import PipelineArtifactElement

            if isinstance(input, PipelineVariable) and input.parent:
                return input.parent
            elif isinstance(input, PipelineArtifactElement) and input.artifact.parent:
                return input.artifact.parent

        input_steps = set(filter(None, [_depend_step(val) for val in values]))

        self._depends = input_steps.union(self._depends)

    @property
    def depends(self):
        return list(self._depends)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @classmethod
    def get_component(cls, identifier, provider, version):
        from pai.pipeline.component import RegisteredComponent

        component = RegisteredComponent.get_by_identifier(
            identifier=identifier, provider=provider, version=version
        )
        return component

    def after(self, *steps):
        if self.parent or any(step for step in steps if step.parent):
            raise ValueError(
                "Not allow operation, pipeline step has been included in a pipeline"
            )
        for step in steps:
            if step not in self._depends:
                self._depends.add(step)

    @property
    def ref_name(self):
        return "pipelines.{}".format(self.name)

    def _convert_spec_to_json(self):
        assigned_inputs = [ipt for ipt in self.inputs if ipt.name in self._assigned]

        repeated_artifact_config = [
            {
                "name": opt.name,
                "value": [None] * opt.count,
            }
            for opt in self.outputs.artifacts
            if opt.repeated and opt.count
        ]

        spec = {
            "arguments": {
                "parameters": [
                    ipt.to_argument()
                    for ipt in assigned_inputs
                    if ipt.variable_category == "parameters"
                ],
                "artifacts": [
                    ipt.to_argument()
                    for ipt in assigned_inputs
                    if ipt.variable_category == "artifacts"
                ]
                + repeated_artifact_config,
            }
        }

        if self._depends:
            spec["dependencies"] = [step.name for step in self.depends]
        return spec

    def to_dict(self):
        metadata = {"name": self.name}
        metadata.update(self.component_ref.to_dict())

        d = {
            "metadata": metadata,
            "spec": self._convert_spec_to_json(),
        }
        return d


class ConditionStep(PipelineStep):
    """Represent a conditional execution step in pipeline."""

    def __init__(self, condition, inputs=None, name=None, depends=None, component=None):
        """Construct a ConditionStep to support conditional execution in pipeline.

        A ConditionStep only execute if condition evaluate to true.

        Args:
            condition (Union[str, ConditionExpr]): Condition expression used to determine
                if the step should be executed, could be ConditionExpr or str.
            inputs (dict): Inputs for the step in dict: key is the component input name, value
                could be the output artifact/parameter from other step, input of the pipeline,
                or actual value for the step.
            name (str): Name of the step in pipeline, must be unique in the pipeline.
            depends (list): A list of PipelineStep which step depends.
            component (ComponentBase): The component used by the constructed step.
        """
        if not isinstance(condition, (ConditionExpr, str)):
            raise ValueError("Not supported condition type: %s" % type(condition))
        elif isinstance(condition, ConditionExpr):
            condition_depends = condition.get_depends_steps()
            depends = list(filter(None, (depends or []) + condition_depends))

        super().__init__(inputs=inputs, name=name, depends=depends, component=component)
        self.condition = condition

    def to_dict(self):
        d = super(ConditionStep, self).to_dict()
        d["spec"]["when"] = (
            self.condition.to_expr()
            if isinstance(self.condition, ConditionExpr)
            else str(self.condition)
        )
        return d


class LoopStep(PipelineStep):
    """Represent a parallel execution step in pipeline."""

    DEFAULT_PARALLELISM_COUNT = 5

    def __init__(
        self,
        items,
        parallelism=None,
        inputs=None,
        name=None,
        depends=None,
        component=None,
    ):
        """Construct a LoopStep to support for-loop execution in pipeline.

        A LoopStep invoke the component in for-loop execution style.

        Args:
            items (Union[str, ConditionExpr]): Condition expression used to determine
                if the step should be executed, could be ConditionExpr or str.
            parallelism (int): Max execution parallelism of the step.
            inputs (dict): Inputs for the step in dict: key is the component input name, value
                could be the output artifact/parameter from other step, input of the pipeline,
                or actual value for the step.
            name (str): Name of the step in pipeline, must be unique in the pipeline.
            depends (list): A list of PipelineStep which step depends.
            component (OperatorBase): The component used by the constructed step.
        """
        if (
            isinstance(items, PipelineParameter)
            and items.parent
            and isinstance(items.parent, PipelineStep)
        ):
            depends = depends or []
            depends.append(items.parent)

        super().__init__(inputs=inputs, name=name, depends=depends, component=component)
        self.items = LoopItems(items)
        self.parallelism = int(parallelism or type(self).DEFAULT_PARALLELISM_COUNT)

    def to_dict(self):
        d = super(LoopStep, self).to_dict()
        d["spec"].update(self.items.to_dict())
        if self.parallelism:
            d["spec"]["parallelism"] = self.parallelism

        return d



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import six

from pai.common.yaml_utils import dump as yaml_dump
from pai.common.yaml_utils import safe_load as yaml_safe_load
from pai.pipeline.component._base import ComponentBase, UnRegisteredComponent
from pai.pipeline.types.spec import load_input_output_spec
from pai.session import get_default_session


class RegisteredComponent(ComponentBase):
    """RegisteredComponent represent the pipeline schema from pipeline/component.

    RegisteredComponent object include the definition of "Workflow" use in PAI pipeline service.
     It could be fetched from remote pipeline service or construct from local Pipeline/Component.
     Saved pipeline component has unique `pipeline_id` which is generated by pipeline service.

    """

    def __init__(self, pipeline_id, manifest=None, workspace_id=None):
        """Template constructor.

        Args:
            manifest: "Workflow" definition of the pipeline.
            pipeline_id: Unique ID for pipeline in PAI service.
            workspace_id: ID of the workspace which the pipeline belongs to.
        """
        if not manifest:
            session = get_default_session()
            manifest = session.pipeline_api.get_schema()["Manifest"]
        if isinstance(manifest, six.string_types):
            manifest = yaml_safe_load(manifest)

        self._manifest = manifest
        self._pipeline_id = pipeline_id
        self._workspace_id = workspace_id

        inputs, outputs = load_input_output_spec(self, manifest["spec"])

        self._identifier = manifest["metadata"]["identifier"]
        self._provider = manifest["metadata"]["provider"]
        self._version = manifest["metadata"]["version"]
        super(RegisteredComponent, self).__init__(
            inputs=inputs,
            outputs=outputs,
        )

    def __repr__(self):
        return "%s:Id=%s,Identifier=%s,Provider=%s,Version=%s" % (
            type(self).__name__,
            self._pipeline_id,
            self.identifier,
            self.provider,
            self.version,
        )

    def __eq__(self, other):
        return isinstance(other, type(self)) and other.pipeline_id == self.pipeline_id

    @property
    def identifier(self):
        return self._identifier

    @property
    def provider(self):
        return self._provider

    @property
    def version(self):
        return self._version

    @property
    def pipeline_id(self):
        """Unique ID of the pipeline in PAI pipeline service.

        Returns:
            str: Unique pipeline ID of the component instance.

        """
        return self._pipeline_id

    @property
    def manifest(self):
        """Pipeline manifest schema.

        Returns:
            dict: Pipeline manifest schema in dict.
        """
        return self._manifest

    @property
    def raw_manifest(self):
        """Pipeline manifest in YAML format

        Returns:
            str: Pipeline manifest.
        """
        return yaml_dump(self._manifest)

    @classmethod
    def get_by_identifier(cls, identifier, provider=None, version="v1"):
        """Get SavedOperator with identifier-provider-version tuple.

        Args:
            identifier (str): Pipeline identifier.
            provider (str): Provider of the Pipeline, account uid of the current session will be used as
              default.
            version (str): Version of the pipeline.

        Returns:
            pai.pipeline.SavedOperator: SavedOperator instance

        """
        session = get_default_session()
        provider = provider or session.provider

        res = session.pipeline_api.get_by_identifier(
            identifier=identifier,
            provider=provider,
            version=version,
        )
        if not res:
            raise ValueError(
                f"Not found the specific pipeline/component: "
                f"identifier={identifier} provider={provider} version={version}"
            )
        pipeline_info = session.pipeline_api.get_schema(pipeline_id=res["PipelineId"])
        if not pipeline_info:
            raise ValueError(
                "Not found pipeline with specific information: identifier={0}, provider={1}, version={2}".format(
                    identifier, provider, version
                )
            )

        return cls(
            manifest=pipeline_info["Manifest"],
            pipeline_id=pipeline_info["PipelineId"],
        )

    @classmethod
    def list(
        cls,
        identifier=None,
        provider=None,
        version=None,
        session=None,
        page_size=10,
        page_number=1,
    ):
        """List the SavedOperator in PAI

        Search the pipeline component available in remote PAI service. The method return a
        generator used to traverse the SavedOperator set match the query condition.

        Args:
            identifier (str): Pipeline identifier filter.
            provider (str): Pipeline provider filter.
            version (str): Pipeline version.
            workspace_id (str): Workspace id of the pipeline.
            session: PAI session.
            page_number (int):
            page_size (int):

        Yields:
              pai.component.SavedOperator: SavedOperator match the query.
        """
        session = session or get_default_session()

        if not provider:
            provider = session.provider

        result = session.pipeline_api.list(
            identifier=identifier,
            provider=provider,
            version=version,
            page_size=page_size,
            page_number=page_number,
        )

        return result.items

    def update(self, component):
        """Update current registered component/pipeline using the manifest of given component/pipeline.

        Args:
            component (Union[UnRegisteredOperator, str, dict]): New pipeline/component spec,
            could be an unregistered component, dict or yaml in str.
        """
        session = get_default_session()

        if isinstance(component, UnRegisteredComponent):
            manifest = component.to_manifest(
                identifier=self.identifier, version=self.version
            )
        elif isinstance(component, str):
            manifest = component
        elif isinstance(component, dict):
            manifest = yaml_dump(component)

        else:
            raise ValueError(
                "Please provider ContainerOperator, Pipeline or Manifest in string to update the registered component."
            )

        session.pipeline_api.update(self._pipeline_id, manifest)

    def delete(self):
        """Delete this registered component/pipeline."""
        get_default_session().pipeline_api.delete(self.pipeline_id)

    @classmethod
    def deserialize(cls, obj_dict):
        manifest, id, workspace_id = (
            obj_dict.get("Manifest"),
            obj_dict["PipelineId"],
            obj_dict.get("WorkspaceId", None),
        )
        return cls(
            workspace_id=workspace_id,
            pipeline_id=id,
            manifest=manifest,
        )

    @classmethod
    def _has_impl(cls, manifest):
        if isinstance(manifest, six.string_types):
            manifest = yaml_safe_load(manifest)

        if "spec" not in manifest:
            return False
        spec = manifest["spec"]
        if "pipelines" not in spec and "container" not in spec:
            return False
        return True

    @classmethod
    def get(cls, pipeline_id, session=None):
        """Get SavedOperator with pipeline_id.

        Args:
            pipeline_id (str): Unique pipeline id.

        Returns:
            pai.pipeline.SavedOperator: SavedOperator instance with the
             specific pipeline_id

        """
        session = session or get_default_session()
        return cls.deserialize(session.pipeline_api.get_schema(pipeline_id=pipeline_id))

    def save(self, identifier=None, version=None):
        raise NotImplementedError("SaveTemplate is not savable.")

    def _submit(self, job_name, args):
        from pai.pipeline.run import PipelineRun

        session = get_default_session()

        run_id = PipelineRun.run(
            job_name,
            args,
            no_confirm_required=True,
            pipeline_id=self._pipeline_id,
            session=session,
        )

        return run_id

    def io_spec_to_dict(self):
        return {
            "inputs": self.inputs.to_dict(),
            "outputs": self.outputs.to_dict(),
        }



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
import threading
from typing import Optional

PAI_LOG_LEVEL = "PAI_LOG_LEVEL"


_lock = threading.Lock()
_default_handler: Optional[logging.Handler] = None

_LOG_LEVEL_MAPPING = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

_default_log_level = logging.WARNING


def _get_default_logging_level():
    """Get the default logging level for the pai library."""
    level_str = os.getenv(PAI_LOG_LEVEL, None)
    if level_str:
        if level_str.lower() in _LOG_LEVEL_MAPPING:
            return _LOG_LEVEL_MAPPING[level_str.lower()]
        else:
            logging.getLogger().warning(
                f"Unknown option PAI_LOG_LEVEL={level_str}, "
                f"has to be one of: { ', '.join(_LOG_LEVEL_MAPPING.keys()) }"
            )
    return _default_log_level


def _get_library_name() -> str:
    return __name__.split(".")[0]


def _get_library_root_logger() -> logging.Logger:
    return logging.getLogger(_get_library_name())


def _configure_library_root_logger() -> None:
    global _default_handler

    with _lock:
        if _default_handler:
            return
        handler = logging.StreamHandler()
        library_root_logger = _get_library_root_logger()
        library_root_logger.addHandler(handler)
        library_root_logger.setLevel(_get_default_logging_level())
        library_root_logger.propagate = False
        _default_handler = handler


def _reset_library_root_logger() -> None:
    global _default_handler

    with _lock:
        if not _default_handler:
            return
        library_root_logger = _get_library_root_logger()
        library_root_logger.removeHandler(_default_handler)
        library_root_logger.setLevel(logging.NOTSET)
        _default_handler = None


def get_log_levels_dict():
    return _LOG_LEVEL_MAPPING


def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Return a logger with the specified module name.
    """

    if name is None:
        name = _get_library_name()

    _configure_library_root_logger()
    return logging.getLogger(name)


def get_log_level() -> int:
    """
    Return the current level for the "pai" module.

    Returns:
        `int`: The logging level.
    """

    _configure_library_root_logger()
    return _get_library_root_logger().getEffectiveLevel()


def set_log_level(verbosity: int) -> None:
    """
    Set the log level for the pai root logger.
    """

    _configure_library_root_logger()
    _get_library_root_logger().setLevel(verbosity)


def set_log_level_info():
    """Set the log level  to the `INFO` level."""
    return set_log_level(logging.INFO)


def set_log_level_warning():
    """Set the log level to the `WARNING` level."""
    return set_log_level(logging.WARNING)


def set_log_level_debug():
    """Set the log level to the `DEBUG` level."""
    return set_log_level(logging.DEBUG)


def set_log_level_error():
    """Set the log level to the `ERROR` level."""
    return set_log_level(logging.ERROR)



#  Copyright 2024 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import posixpath
import time
import typing
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_pascal
from Tea.exceptions import TeaException

from ..api.base import PaginatedResult
from ..common.consts import StoragePathCategory
from ..common.logging import get_logger
from ..common.oss_utils import OssUriObj, is_oss_uri, upload
from ..common.utils import (
    is_dataset_id,
    is_filesystem_uri,
    is_odps_table_uri,
    name_from_base,
    print_table,
    random_str,
    retry,
    to_plain_text,
)
from ..exception import UnexpectedStatusException
from ..session import Session, get_default_session

if typing.TYPE_CHECKING:
    from ..estimator import FileSystemInputBase

logger = get_logger(__name__)


def as_oss_dir_uri(uri: str):
    return uri if uri.endswith("/") else uri + "/"


DEFAULT_OUTPUT_MODEL_CHANNEL_NAME = "model"
DEFAULT_CHECKPOINT_CHANNEL_NAME = "checkpoints"
DEFAULT_TENSORBOARD_CHANNEL_NAME = "tensorboard"


class SpotStrategy(str, Enum):
    SpotWithPriceLimit = "SpotWithPriceLimit"
    SpotAsPriceGo = "SpotAsPriceGo"

    def __repr__(self):
        return self.value


class ResourceType(str, Enum):
    Lingjun = "Lingjun"
    General = "General"


class BaseAPIModel(BaseModel):

    model_config = ConfigDict(
        alias_generator=to_pascal,
        populate_by_name=True,
    )

    def model_dump(self, **kwargs) -> Dict[str, Any]:
        kwargs.update({"by_alias": True, "exclude_none": True})
        return super().model_dump(**kwargs)

    def to_dict(self):
        return self.model_dump()


class TrainingJobStatus(object):
    CreateFailed = "CreateFailed"
    InitializeFailed = "InitializeFailed"
    Succeed = "Succeed"
    Failed = "Failed"
    Terminated = "Terminated"
    Creating = "Creating"
    Created = "Created"
    Initializing = "Initializing"
    Submitted = "Submitted"
    Running = "Running"

    @classmethod
    def completed_status(cls):
        return [
            cls.InitializeFailed,
            cls.Succeed,
            cls.Failed,
            cls.Terminated,
        ]

    @classmethod
    def failed_status(cls):
        return [
            cls.InitializeFailed,
            cls.Failed,
            cls.CreateFailed,
        ]


class UserVpcConfig(BaseAPIModel):
    """UserVpcConfig represents the VPC configuration for the training job instance."""

    vpc_id: str = Field(
        ...,
        description="Specifies the ID of the VPC that training job instance connects to.",
    )
    security_group_id: str = Field(
        ...,
        description="The ID of the security group that training job instances belong to.",
    )
    switch_id: Optional[str] = Field(
        None,
        description="The ID of the vSwitch to which the instance belongs. Defaults to None.",
    )
    extended_cidrs: Optional[List[str]] = Field(
        None,
        description="The CIDR blocks configured for the ENI of the training job instance. "
        "If it is not specified, the CIDR block will be configured as the same as the VPC "
        "network segmentation, which means that the training job instance can access all "
        "resources in the VPC. Defaults to None.",
    )


class ExperimentConfig(BaseAPIModel):
    """ExperimentConfig is used to configure the experiment to which the job belongs."""

    experiment_id: str = Field(
        ...,
        description="Specifies the ID of the experiment that training job instance belongs to.",
    )


class OssLocation(BaseAPIModel):
    """OSS location."""

    bucket: str = Field(..., description="OSS bucket name.")
    key: str = Field(..., description="Object key in the OSS bucket.")
    endpoint: Optional[str] = Field(None, description="OSS service endpoint URL.")


class CodeDir(BaseAPIModel):
    """Source code location"""

    location_value: Union[OssLocation, Dict[str, Any]] = Field(
        ..., description="Location of the code directory."
    )
    location_type: str = Field(
        ..., description="Type of the code directory location, e.g., OSS."
    )


# HyperParameter
class HyperParameter(BaseAPIModel):
    """A hyperparameter for a training job."""

    value: str = Field(..., description="Value of the hyperparameter.")
    name: str = Field(..., description="Name of the hyperparameter.")


class InstanceSpec(BaseAPIModel):
    """Instance resource configuration"""

    memory: str = Field(..., description="Memory allocation for the instance.")
    cpu: str = Field(..., alias="CPU", description="CPU allocation for the instance.")
    gpu: str = Field(..., alias="GPU", description="GPU allocation for the instance.")
    shared_memory: Optional[str] = Field(
        None, description="Shared memory allocation, if applicable."
    )


class ComputeResource(BaseAPIModel):
    """Compute Resource Configuration."""

    ecs_count: Optional[int] = Field(None, description="Number of ECS instances.")
    ecs_spec: Optional[str] = Field(None, description="Specification of ECS instances.")
    instance_count: Optional[int] = Field(None, description="Number of instances.")
    instance_spec: Optional[InstanceSpec] = Field(
        None, description="Specification for instances."
    )


# URI Input and Output
class UriInput(BaseAPIModel):
    """URI Input for a training job."""

    name: str = Field(..., description="Name of the input.")
    input_uri: str = Field(..., description="URI of the input data.")


class UriOutput(BaseAPIModel):
    """URI Output for a training job."""

    name: str = Field(..., description="Name of the output.")
    output_uri: str = Field(..., description="URI of the output data.")


class DatasetConfig(BaseAPIModel):
    """Dataset Configuration"""

    dataset_id: str = Field(..., description="Unique ID of the dataset.")
    name: Optional[str] = Field(None, description="Name of the dataset.")
    dataset_name: Optional[str] = Field(
        None, description="Alternative name of the dataset."
    )


class Channel(BaseAPIModel):
    """Channel Configuration."""

    name: str = Field(..., description="Name of the channel.")
    description: Optional[str] = Field(None, description="Description of the channel.")
    required: Optional[bool] = Field(
        None, description="Indicates if the channel is required."
    )
    supported_channel_types: Optional[List[str]] = Field(
        None, description="Supported types for this channel."
    )
    properties: Optional[Dict[str, Any]] = Field(
        None, description="Additional properties of the channel."
    )


# HyperParameter Definition
class HyperParameterDefinition(BaseAPIModel):
    """HyerParameter Definition."""

    name: str = Field(..., description="Name of the hyperparameter.")
    type: Optional[str] = Field(None, description="Type of the hyperparameter.")
    default_value: Optional[str] = Field(
        None, description="Default value of the hyperparameter."
    )
    description: Optional[str] = Field(
        None, description="Description of the hyperparameter."
    )
    required: bool = Field(
        False, description="Indicates if the hyperparameter is required."
    )


class SchedulerConfig(BaseAPIModel):
    max_running_time_in_seconds: Optional[int] = None


class MetricDefinition(BaseAPIModel):
    description: Optional[str] = Field(None, description="Description of the metric.")
    name: str = Field(..., description="Name of the metric.")
    regex: str = Field(
        ..., description="Regular expression used for capturing the metric."
    )


class AlgorithmSpec(BaseAPIModel):
    """Algorithm Specification."""

    command: List[str] = Field(..., description="Command to run the training job.")
    image: str = Field(..., description="Docker image for the training job.")
    supported_channel_types: List[str] = Field(default_factory=list)
    output_channels: List[Channel] = Field(
        default_factory=list, description="Output channels."
    )
    input_channels: List[Channel] = Field(
        default_factory=list, description="Input channels."
    )
    supports_distributed_training: Optional[bool] = Field(
        True, description="Whether the algorithm supports distributed training."
    )
    supported_instance_types: Optional[List[str]] = Field(
        None, description="Supported instance types."
    )
    metric_definitions: Optional[List[MetricDefinition]] = Field(
        None, description="Metric definitions."
    )
    hyperparameter_definitions: List[HyperParameterDefinition] = Field(
        default_factory=list,
        alias="HyperParameters",
        description="Hyperparameter definitions.",
    )
    job_type: str = Field(default="PyTorchJob")
    code_dir: Optional[CodeDir] = Field(None, description="Source code location.")
    customization: Optional[Dict[str, Any]] = Field(
        None, description="Whether the algorithm supports customize code."
    )


class ModelRecipeSpec(BaseAPIModel):
    compute_resource: Optional[ComputeResource] = None
    hyperparameters: List[HyperParameter] = Field(
        default_factory=list, alias="HyperParameters"
    )
    inputs: List[Union[UriInput, DatasetConfig]] = Field(
        default_factory=list, alias="InputChannels"
    )
    scheduler: Optional[SchedulerConfig] = None
    supported_instance_types: Optional[List[str]] = None
    algorithm_spec: Optional[AlgorithmSpec] = None
    algorithm_version: Optional[str] = None
    algorithm_provider: Optional[str] = None
    algorithm_name: Optional[str] = None
    environments: Optional[Dict[str, str]] = None
    requirements: Optional[List[str]] = None


class SpotSpec(BaseAPIModel):
    spot_strategy: SpotStrategy = Field(
        ...,
        description="Spot instance strategy, support 'SpotWithPriceLimit', 'SpotAsPriceGo'",
    )
    spot_discount_limit: Optional[float] = Field(
        None,
        description="Spot instance discount limit, maximum 2 decimal places, "
        "required when spot_strategy is 'SpotWithPriceLimit'."
        "For example, 0.5 means 50% off the original price.",
    )


class TrainingJob(BaseAPIModel):
    """TrainingJob represents a training job in the PAI service."""

    algorithm_id: Optional[str] = None
    algorithm_name: Optional[str] = None
    algorithm_provider: Optional[str] = None
    algorithm_version: Optional[str] = None
    algorithm_spec: Optional[AlgorithmSpec] = None
    compute_resource: Optional[ComputeResource] = None
    scheduler: Optional[SchedulerConfig] = None
    experiment_config: Optional[Dict[str, Any]] = None
    inputs: List[Union[UriInput, DatasetConfig]] = Field(
        default=list, alias="InputChannels"
    )
    outputs: List[Union[UriOutput, DatasetConfig]] = Field(
        default=list, alias="OutputChannels"
    )
    hyperparameters: List[HyperParameter] = Field(
        default_factory=list, alias="HyperParameters"
    )
    labels: Optional[List[Dict[str, str]]] = Field(default_factory=list)
    training_job_description: Optional[str] = None
    training_job_id: Optional[str] = None
    training_job_name: Optional[str] = None
    workspace_id: Optional[str] = None
    training_job_url: Optional[str] = None
    status: Optional[str] = None
    reason_code: Optional[str] = None
    reason_message: Optional[str] = None

    def __hash__(self):
        return hash(self.training_job_id)

    def __eq__(self, other: "TrainingJob"):
        return (
            isinstance(other, TrainingJob)
            and self.training_job_id == other.training_job_id
        )

    @property
    def id(self):
        return self.training_job_id

    @classmethod
    def get(cls, training_job_id, session: Session = None) -> "TrainingJob":
        session = session or get_default_session()
        res = session.training_job_api.get(training_job_id=training_job_id)
        return cls.model_validate(res)

    @classmethod
    def list(
        cls,
        status: Optional[str] = None,
        session: Optional[Session] = None,
        page_size: int = 50,
        page_number: int = 1,
    ):
        session = session or get_default_session()
        res = session.training_job_api.list(
            status=status, page_size=page_size, page_number=page_number
        )
        return [cls.model_validate(item) for item in res.items]

    def output_path(self, channel_name="model"):
        for output_channel in self.outputs:
            if output_channel.name == channel_name:
                return output_channel.output_uri
        raise RuntimeError(
            f"Output channel is not specified: channel_name={channel_name}"
        )

    @property
    def console_uri(self):
        if not self.training_job_id:
            raise ValueError("The TrainingJob is not submitted")

        return self.training_job_url

    def wait(self, interval: int = 5, show_logs: bool = True):
        session = get_default_session()
        self._refresh_status()

        if show_logs:
            job_log_printer = _TrainingJobLogPrinter(
                training_job_id=self.training_job_id, page_size=20, session=session
            )
            job_log_printer.start()
        else:
            job_log_printer = None
        try:
            while not self.is_completed():
                time.sleep(interval)
        finally:
            if job_log_printer:
                job_log_printer.stop(wait=True)

        self._on_job_completed()

    def _on_job_completed(self):
        # Print an empty line to separate the training job logs and the following logs
        print()
        if self.status == TrainingJobStatus.Succeed:
            print(
                f"Training job ({self.training_job_id}) succeeded, you can check the"
                f" logs/metrics/output in  the console:\n{self.console_uri}"
            )
        elif self.status == TrainingJobStatus.Terminated:
            print(
                f"Training job is ended with status {self.status}: "
                f"reason_code={self.reason_code}, reason_message={self.reason_message}."
                f"Check the training job in the console:\n{self.console_uri}"
            )
        elif self.status in TrainingJobStatus.failed_status():
            print(
                f"Training job ({self.training_job_id}) failed, please check the logs"
                f" in the console: \n{self.console_uri}"
            )

            message = f"TrainingJob failed: name={self.training_job_name}, "
            f"training_job_id={self.training_job_id}, "
            f"reason_code={self.reason_code}, status={self.status}, "
            f"reason_message={self.reason_message}"

            raise UnexpectedStatusException(message=message, status=self.status)

    def _refresh_status(self):
        """Reload the training job from the PAI Service,"""
        session = get_default_session()
        training_job = type(self).model_validate(
            session.training_job_api.get(training_job_id=self.training_job_id)
        )
        self.status = training_job.status

    def is_succeeded(self):
        """Return True if the training job is succeeded"""
        self._refresh_status()
        return self.status == TrainingJobStatus.Succeed

    @retry(wait_secs=10)
    def is_completed(self):
        """Return True if the training job is completed, including failed status"""
        if self.status in TrainingJobStatus.completed_status():
            return True
        self._refresh_status()

        return self.status in TrainingJobStatus.completed_status()


class _TrainingJobLogPrinter(object):
    """A class used to print logs for a training job"""

    executor = ThreadPoolExecutor(5)

    def __init__(
        self, training_job_id: str, page_size=10, session: Optional[Session] = None
    ):
        self.training_job_id = training_job_id
        self.session = session
        self.page_size = page_size
        self._future = None
        self._stop = False

    def _list_logs_api(self, page_number: int = 1):
        try:
            res = self.session.training_job_api.list_logs(
                self.training_job_id,
                page_number=page_number,
                page_size=self.page_size,
            )
            return res
        except TeaException as e:
            # hack: Backend service may raise an exception when the training job
            # instance is not found.
            if e.code == "TRAINING_JOB_INSTANCE_NOT_FOUND":
                return PaginatedResult(items=[], total_count=0)
            else:
                raise e

    def _list_logs(self):
        page_number, page_offset = 1, 0
        # print training job logs.
        while not self._stop:
            res = self._list_logs_api(page_number=page_number)
            # 1. move to next page
            if len(res.items) == self.page_size:
                # print new logs starting from page_offset
                self._print_logs(logs=res.items[page_offset:])
                page_number += 1
                page_offset = 0
            # 2. stay at the current page.
            else:
                if len(res.items) > page_offset:
                    # print new logs starting from page_offset
                    self._print_logs(logs=res.items[page_offset:])
                    page_offset = len(res.items)
                time.sleep(1)

        # When _stop is True, wait and print remaining logs.
        time.sleep(10)
        while True:
            res = self._list_logs_api(page_number=page_number)
            # There maybe more logs in the next page
            if len(res.items) == self.page_size:
                self._print_logs(logs=res.items[page_offset:])
                page_number += 1
                page_offset = 0
            # No more logs in the next page.
            else:
                if len(res.items) > page_offset:
                    self._print_logs(logs=res.items[page_offset:])
                break

    def _print_logs(self, logs: List[str]):
        for log in logs:
            print(log)

    def start(self):
        if self._future:
            raise ValueError("The training job log printer is already started")
        self._stop = False
        self._future = self.executor.submit(self._list_logs)

    def stop(self, wait: bool = True):
        self._stop = True
        if self._future:
            self._future.result()


class _TrainingJobSubmitter(object):
    """A class used to submit a training job to the PAI service."""

    def __init__(
        self,
        base_job_name: Optional[str] = None,
        output_path: Optional[str] = None,
        experiment_config: Optional[ExperimentConfig] = None,
        user_vpc_config: Optional[UserVpcConfig] = None,
        max_run_time: Optional[int] = None,
        instance_type: Optional[str] = None,
        instance_spec: Optional[Dict] = None,
        instance_count: Optional[int] = None,
        resource_id: Optional[Dict] = None,
        resource_type: Optional[Union[str, ResourceType]] = None,
        spot_spec: Optional[SpotSpec] = None,
        environments: Optional[Dict] = None,
        requirements: Optional[List[str]] = None,
        labels: Optional[Dict[str, str]] = None,
        settings: Optional[Dict[str, Any]] = None,
    ):
        self.session = get_default_session()
        self._training_jobs = []
        self.base_job_name = base_job_name or type(self).__name__.lower()
        self.output_path = output_path
        self.user_vpc_config = user_vpc_config
        self.spot_spec = spot_spec
        self.experiment_config = experiment_config
        self.max_run_time = max_run_time
        self.instance_type = instance_type
        self.instance_spec = instance_spec
        self.instance_count = instance_count or 1
        self.resource_id = resource_id
        self.resource_type = ResourceType(resource_type) if resource_type else None
        self.environments = environments
        self.requirements = requirements
        self.settings = settings
        self.labels = labels

    def wait(self, interval: int = 5, show_logs: bool = True, all_jobs: bool = False):
        """Block until the jobs is completed.

        Args:
            interval(int): Interval to reload job status
            show_logs(bool): Specifies whether to fetch and print the logs produced by
                the job.
            all_jobs(bool): Wait latest job or wait all jobs in processor, show_logs disabled while
                wait all jobs.

        Raises:
            RuntimeError: If no job is submitted.

        """
        if all_jobs:
            if not self._training_jobs:
                raise RuntimeError("Could not find any submitted job.")
            remains = set(self._training_jobs)
            while remains:
                for job in self._training_jobs:
                    if job in remains and job.is_completed():
                        remains.remove(job)

                time.sleep(interval)
            self._generate_jobs_report()
        else:
            latest_job = self.latest_job
            if not latest_job:
                raise RuntimeError("Could not find a submitted job.")
            latest_job.wait(interval=interval, show_logs=show_logs)
            return latest_job

    def _generate_jobs_report(self):
        """Generate current jobs report and output to stdout"""
        print(f"Jobs status report, total jobs count: {len(self._training_jobs)}")
        rows = []
        headers = ["JobName", "JobID", "Status"]
        for job in self._training_jobs:
            rows.append([job.training_job_name, job.id, job.status])
        print_table(headers, rows)

    def job_name(self, job_name: Optional[str] = None):
        if job_name:
            return job_name
        sep = "-"
        base_name = self.base_job_name
        return name_from_base(base_name, sep)

    def build_inputs(
        self,
        inputs: Dict[str, Any],
        input_channels: List[Channel],
        default_inputs: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, str]]:
        res = []
        inputs = inputs or dict()
        input_channels = input_channels or []
        default_inputs = default_inputs or {}

        inputs = {**default_inputs, **inputs}
        requires = {ch.name for ch in input_channels if ch.required} - set(
            inputs.keys()
        )
        if requires:
            raise ValueError(
                "Required input channels are not provided: {}".format(
                    ",".join(requires)
                )
            )
        for name, item in inputs.items():
            input_config = self._get_input_config(name, item)
            res.append(input_config.model_dump())

        return res

    @staticmethod
    def _default_training_output_channels() -> List[Channel]:
        channels = [
            Channel(
                name=DEFAULT_OUTPUT_MODEL_CHANNEL_NAME,
                description="Training output models",
                required=True,
            ),
            Channel(
                name=DEFAULT_CHECKPOINT_CHANNEL_NAME,
                description="Training checkpoints channel",
                required=False,
            ),
            Channel(
                name=DEFAULT_TENSORBOARD_CHANNEL_NAME,
                properties={"ossAppendable": "true"},
                description="TensorBoard logs channel",
                required=False,
            ),
        ]

        return channels

    def _training_job_base_output(self, job_name):
        job_name = to_plain_text(job_name)
        if self.output_path:
            if not is_oss_uri(self.output_path):
                raise ValueError("Output path should be an OSS path.")
            return os.path.join(self.output_path, f"{job_name}_{random_str(6)}")

        session = get_default_session()
        bucket_name = session.oss_bucket.bucket_name
        storage_path = session.get_storage_path_by_category(
            StoragePathCategory.TrainingJob,
            f"{to_plain_text(job_name)}_{random_str(6)}",
        )
        base_output_path = f"oss://{bucket_name}/{storage_path}"
        return base_output_path

    def build_outputs(
        self,
        job_name: str,
        output_channels: List[Channel],
        outputs: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, str]]:
        base_output_path = self._training_job_base_output(job_name)
        res = []
        outputs = outputs or dict()

        for ch in output_channels:
            if ch.name in outputs:
                output = self._get_output_config(name=ch.name, item=outputs[ch.name])
            else:
                output_uri = as_oss_dir_uri(posixpath.join(base_output_path, ch.name))
                output = UriOutput(name=ch.name, output_uri=output_uri)
            res.append(output)

        extra_outputs = set(outputs.keys()) - {ch.name for ch in output_channels}

        for name in extra_outputs:
            output = self._get_output_config(
                name=name,
                item=outputs[name],
            )
            res.append(output)

        return [item.model_dump() for item in res]

    # TODO: get arguments, such as VPCConfig, instance_type etc, from self instance.
    def _submit(
        self,
        job_name: str,
        algorithm_spec: Optional[AlgorithmSpec] = None,
        algorithm_name: Optional[str] = None,
        algorithm_version: Optional[str] = None,
        algorithm_provider: Optional[str] = None,
        instance_count: int = 1,
        instance_type: Optional[str] = None,
        instance_spec: Optional[InstanceSpec] = None,
        resource_id: Optional[str] = None,
        inputs: Optional[List[Dict[str, Any]]] = None,
        outputs: Optional[List[Dict[str, Any]]] = None,
        hyperparameters: Optional[Dict[str, str]] = None,
        max_run_time: Optional[int] = None,
        environments: Optional[Dict[str, str]] = None,
        user_vpc_config: Optional[Dict[str, str]] = None,
        requirements: Optional[List[str]] = None,
        experiment_config: Optional[Dict[str, Any]] = None,
        labels: Optional[Dict[str, str]] = None,
        wait: bool = True,
        show_logs: bool = False,
    ):
        session = get_default_session()

        if not self.resource_type or self.resource_type == ResourceType.General:
            resource_type = None
        else:
            resource_type = self.resource_type.value

        if self.spot_spec:
            spot_spec = {
                "SpotStrategy": self.spot_spec.spot_strategy.value,
            }
            if self.spot_spec.spot_discount_limit:
                spot_spec["SpotDiscountLimit"] = self.spot_spec.spot_discount_limit
        else:
            spot_spec = None

        # user vpc
        if self.user_vpc_config:
            user_vpc_config = {
                "VpcId": self.user_vpc_config.vpc_id,
                "SecurityGroupId": self.user_vpc_config.security_group_id,
            }
        else:
            user_vpc_config = None

        training_job_id = session.training_job_api.create(
            instance_count=instance_count,
            instance_spec=instance_spec.model_dump() if instance_spec else None,
            algorithm_name=algorithm_name,
            algorithm_provider=algorithm_provider,
            experiment_config=(
                experiment_config.model_dump()
                if experiment_config and isinstance(experiment_config, ExperimentConfig)
                else experiment_config
            ),
            spot_spec=spot_spec,
            algorithm_version=algorithm_version,
            instance_type=instance_type,
            resource_id=resource_id,
            resource_type=resource_type,
            job_name=job_name,
            hyperparameters=hyperparameters,
            max_running_in_seconds=max_run_time,
            input_channels=inputs,
            output_channels=outputs,
            algorithm_spec=algorithm_spec.model_dump() if algorithm_spec else None,
            requirements=requirements,
            user_vpc_config=user_vpc_config,
            labels=labels,
            environments=environments,
            settings=self.settings,
        )
        training_job = TrainingJob.get(training_job_id)
        self._training_jobs.append(training_job)
        print(
            f"View the job detail by accessing the console URI: {training_job.console_uri}"
        )
        if wait:
            training_job.wait(show_logs=show_logs)
        return training_job

    @classmethod
    def _get_input_config(
        cls, name: str, item: Union[str, "FileSystemInputBase", DatasetConfig]
    ) -> Union[UriInput, DatasetConfig]:
        """Get input uri for training_job from given input."""
        from pai.estimator import FileSystemInputBase

        if not isinstance(item, (str, FileSystemInputBase, DatasetConfig)):
            raise ValueError(f"Input data of type {type(item)} is not supported.")

        if isinstance(item, FileSystemInputBase):
            input_ = UriInput(
                name=name,
                input_uri=item.to_input_uri(),
            )
        elif isinstance(item, DatasetConfig):
            input_ = DatasetConfig(
                name=name,
                dataset_id=item.dataset_id,
            )
        elif is_oss_uri(item) or is_filesystem_uri(item) or is_odps_table_uri(item):
            input_ = UriInput(
                name=name,
                input_uri=item,
            )
        elif isinstance(item, str):
            if os.path.exists(item):
                store_path = Session.get_storage_path_by_category(
                    StoragePathCategory.InputData
                )
                input_ = UriInput(name=name, input_uri=upload(item, store_path))
            else:
                raise ValueError("Invalid input data path, file not found: {item}.")
        else:
            raise ValueError(
                f"Invalid input data, supported inputs are OSS, NAS, MaxCompute "
                f"table or local path: {type(item)}."
            )
        return input_

    @classmethod
    def _get_output_config(
        cls, name: str, item: str
    ) -> Union[UriOutput, DatasetConfig]:
        from pai.estimator import FileSystemInputBase

        if not isinstance(item, (str, FileSystemInputBase, DatasetConfig)):
            raise ValueError(f"Output data of type {type(item)} is not supported.")

        if isinstance(item, FileSystemInputBase):
            output = UriOutput(
                name=name,
                output_uri=item.to_input_uri(),
            )
        elif isinstance(item, DatasetConfig):
            output = DatasetConfig(name=name, dataset_id=item.dataset_id)
        elif is_oss_uri(item) or is_filesystem_uri(item) or is_odps_table_uri(item):
            output = UriOutput(
                name=name,
                output_uri=as_oss_dir_uri(item),
            )
        else:
            raise ValueError(
                "Invalid output data, supported outputs are OSS, NAS, MaxCompute "
            )

        return output

    @property
    def latest_job(self) -> "TrainingJob":
        return self._training_jobs[-1] if self._training_jobs else None

    def _build_code_input(
        self, job_name: str, source_dir: Optional[str], code_dest: Optional[str] = None
    ) -> Optional[CodeDir]:
        """Upload source files to OSS and return the code input for training job."""
        if not source_dir:
            return
        if is_oss_uri(source_dir):
            code_uri = source_dir
        elif not os.path.exists(source_dir):
            raise ValueError(f"Source directory {source_dir} does not exist.")
        else:
            code_dest = code_dest or self.session.get_storage_path_by_category(
                StoragePathCategory.TrainingSrc, to_plain_text(job_name)
            )
            code_uri = upload(
                source_path=source_dir,
                oss_path=code_dest,
                bucket=self.session.oss_bucket,
            )
        oss_uri_obj = OssUriObj(uri=self.session.patch_oss_endpoint(code_uri))
        code_dir = CodeDir(
            location_type="oss",
            location_value=OssLocation(
                bucket=oss_uri_obj.bucket_name,
                key=oss_uri_obj.object_key,
                endpoint=oss_uri_obj.endpoint,
            ),
        )

        return code_dir



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Union

from ..libs.alibabacloud_aiworkspace20210204.models import (
    ListImageLabelsRequest,
    ListImageLabelsResponseBody,
    ListImagesRequest,
    ListImagesResponseBody,
)
from .base import PaginatedResult, ServiceName, WorkspaceScopedResourceAPI

SUPPORTED_IMAGE_FRAMEWORKS = [
    "DeepRec",
    "DeepSpeed",
    "Megatron-LM",
    "ModelScope",
    "Nemo",
    "OneFlow",
    "PyTorch",
    "TensorFlow",
    "Transformers",
    "XGBoost",
]

SUPPORTED_IMAGE_LANGUAGES = [
    "python",  # TODO: "Python"
]


class ImageLabel(object):
    """Image Label Class."""

    # Unofficial Image Label
    UNOFFICIAL_LABEL = "system.official=false"
    # Official Image Label
    OFFICIAL_LABEL = "system.official=true"

    # PAI Image Label
    PROVIDER_PAI_LABEL = "system.origin=PAI"
    # Community Image Label
    PROVIDER_COMMUNITY_LABEL = "system.origin=Community"

    # DLC Image Label: for training
    DLC_LABEL = "system.supported.dlc=true"
    # EAS Image Label: for inference
    EAS_LABEL = "system.supported.eas=true"
    # DSW Image Label: for develop
    DSW_LABEL = "system.supported.dsw=true"

    # Accelerator: Use GPU
    DEVICE_TYPE_GPU = "system.chipType=GPU"
    DEVICE_TYPE_CPU = "system.chipType=CPU"

    # Python Version
    # TODO: delete this label key
    PYTHON_VERSION = "system.pythonVersion"

    @staticmethod
    def framework_version(
        framework: str,
        version: str,
    ):
        """Create a label for filtering images that support specific framework version.

        Args:
            framework (str): framework name, which is case sensitive.
            version (str): framework version. If version is '*', it will match all
                versions.

        Returns:
            str: framework version label string.

        Raises:
            ValueError: If the framework is not supported.
        """
        if framework not in SUPPORTED_IMAGE_FRAMEWORKS:
            raise ValueError(
                f"Unsupported framework: {framework}. Current supported frameworks are:"
                f" {SUPPORTED_IMAGE_FRAMEWORKS}"
            )
        return f"system.framework.{framework}={version}"

    @staticmethod
    def language_version(
        language: str,
        version: str,
    ):
        """Create a label for filtering images that support specific language version.

        Args:
            language (str): language name, which is case sensitive.
            version (str): language version. If version is '*', it will match all
                versions.

        Returns:
            str: language version label string.

        Raises:
            ValueError: If the language is not supported.
        """
        if language not in SUPPORTED_IMAGE_LANGUAGES:
            raise ValueError(
                f"Unsupported language: {language}. Current supported languages are:"
                f" {SUPPORTED_IMAGE_LANGUAGES}"
            )
        # TODO: "system.language.{language}={version}"
        return f"system.{language}Version={version}"


class ImageAPI(WorkspaceScopedResourceAPI):
    """Class which provide API to operate CodeSource resource."""

    BACKEND_SERVICE_NAME = ServiceName.PAI_WORKSPACE

    _list_method = "list_images_with_options"
    _create_method = "create_image_with_options"
    _list_labels_method = "list_image_labels_with_options"

    def list(
        self,
        labels: Union[Dict[str, Any], List[str]] = None,
        name: str = None,
        order: str = "DESC",
        page_number: int = 1,
        page_size: int = 50,
        parent_user_id: str = None,
        query: str = None,
        sort_by: str = None,
        user_id: str = None,
        verbose: bool = False,
        **kwargs,
    ) -> PaginatedResult:
        """List image resources."""
        workspace_id = kwargs.pop("workspace_id", None)
        if isinstance(labels, dict):
            labels = ",".join(["{}={}".format(k, v) for k, v in labels.items()])
        elif isinstance(labels, list):
            labels = ",".join([item for item in labels])

        req = ListImagesRequest(
            labels=labels,
            name=name,
            order=order,
            page_number=page_number,
            page_size=page_size,
            parent_user_id=parent_user_id,
            query=query,
            sort_by=sort_by,
            user_id=user_id,
            verbose=verbose,
            workspace_id=workspace_id,
        )

        return self._list(request=req)

    def _list(self, request) -> PaginatedResult:
        resp: ListImagesResponseBody = self._do_request(
            self._list_method, request=request
        )

        return self.make_paginated_result(resp)

    def list_labels(
        self,
        image_id: str = None,
        label_filter: Union[Dict[str, Any], List[str]] = None,
        label_keys: str = None,
        region: str = None,
        **kwargs,
    ) -> dict:
        workspace_id = kwargs.pop("workspace_id", None)
        if isinstance(label_filter, dict):
            label_filter = ",".join(
                ["{}={}".format(k, v) for k, v in label_filter.items()]
            )
        elif isinstance(label_filter, list):
            label_filter = ",".join([item for item in label_filter])

        request = ListImageLabelsRequest(
            image_id=image_id,
            label_filter=label_filter,
            label_keys=label_keys,
            region=region,
            workspace_id=workspace_id,
        )
        resp: ListImageLabelsResponseBody = self._do_request(
            method_=self._list_labels_method, request=request
        )
        return resp.to_map()["Labels"]



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from typing import Any, Dict, List, Optional

from .api.image import SUPPORTED_IMAGE_FRAMEWORKS, ImageLabel
from .common.logging import get_logger
from .common.utils import make_list_resource_iterator, to_semantic_version
from .session import Session, get_default_session

logger = get_logger(__name__)


_NORMALIZED_FRAMEWORK_NAMES = {
    name.lower(): name for name in SUPPORTED_IMAGE_FRAMEWORKS
}

# Regex expression pattern for PAI Docker Image Tag.
_PAI_IMAGE_TAG_PATTERN_TRAINING = re.compile(
    r"([\w._-]+)-(gpu|cpu|mkl-cpu)-(py\d+)(?:-(cu\d+))?-([\S]+)"
)

_PAI_IMAGE_TAG_PATTERN_INFERENCE = re.compile(
    r"([\w._-]+)-(py\d+)(?:-(gpu|cpu|mkl-cpu))?(?:-(cu\d+))?-([\S]+)"
)

# Regex expression pattern for PAI Docker Image URI.
_PAI_IMAGE_URI_PATTERN = re.compile(r"([\S]+)/([\S]+)/([\S]+):([\S]+)")


class ImageInfo(object):
    """This class represents information for an image provided by PAI.

    Args:
        image_name (str): The name of the image.
        image_uri (str): The URI of the image.
        framework_name (str): The name of the framework installed in the image.
        framework_version (str, optional): The version of the framework (Default None).
        image_scope (str): The scope of the image, could be 'training', 'inference' or
            'develop'.
        accelerator_type (str, optional): The type of accelerator. Defaults to None.
        python_version (str, optional): The version of Python. Defaults to None.
    """

    def __repr__(self):
        return (
            "{}(framework_name={}: framework_version={}: image_scope={}: "
            "accelerator_type={}: py_version={})".format(
                self.__class__.__name__,
                self.framework_name,
                self.framework_version,
                self.image_scope,
                self.accelerator_type,
                self.python_version,
            )
        )

    def __init__(
        self,
        image_name: str,
        image_uri: str,
        framework_name: str,
        image_scope: str,
        framework_version: str = None,
        accelerator_type: Optional[str] = None,
        python_version: Optional[str] = None,
    ):
        self.image_name = image_name
        self.image_uri = image_uri
        self.framework_name = framework_name
        self.framework_version = framework_version
        self.accelerator_type = accelerator_type
        self.python_version = python_version
        self.image_scope = image_scope


class ImageScope(object):
    """Class containing constants that indicate the purpose of an image."""

    TRAINING = "training"
    """Indicates the image is used for submitting a training job."""

    INFERENCE = "inference"
    """Indicates the image is used for creating a prediction service."""

    DEVELOP = "develop"
    """Indicates the image is used for running in DSW."""

    _SCOPE_IMAGE_LABEL_MAPPING = {
        TRAINING: ImageLabel.DLC_LABEL,
        INFERENCE: ImageLabel.EAS_LABEL,
        DEVELOP: ImageLabel.DSW_LABEL,
    }

    @classmethod
    def to_image_label(cls, scope: str):
        cls._validate(scope)

        return cls._SCOPE_IMAGE_LABEL_MAPPING.get(scope.lower())

    @classmethod
    def _validate(cls, scope: str):
        items = cls._SCOPE_IMAGE_LABEL_MAPPING.keys()
        if scope.lower() not in items:
            raise ValueError(f"Not supported image scope:  {scope}")


def _make_image_info(
    image_obj: Dict[str, Any],
    image_scope: str,
) -> Optional[ImageInfo]:
    """Make a ImageProperties object by parsing the image_uri."""

    labels = {lb["Key"]: lb["Value"] for lb in image_obj["Labels"]}
    image_uri = image_obj["ImageUri"]
    match = _PAI_IMAGE_URI_PATTERN.match(image_uri)
    if not match:
        # ignore if image uri is not recognized
        logger.debug(
            "Could not recognize the given image uri, ignore the image:"
            f" image_uri={image_uri}"
        )
        return
    host, namespace, repo_name, tag = match.groups()

    tag_match = _PAI_IMAGE_TAG_PATTERN_TRAINING.match(
        tag
    ) or _PAI_IMAGE_TAG_PATTERN_INFERENCE.match(tag)
    if not tag_match:
        # ignore if image tag is not recognized
        logger.debug(
            f"Could not recognize the given image tag, ignore the image:"
            f" image_uri={image_uri}."
        )
        fw_version, cpu_or_gpu, _, cuda_version, os_version = [None] * 5
    else:
        (
            fw_version,
            cpu_or_gpu,
            _,
            cuda_version,
            os_version,
        ) = tag_match.groups()

    # use image label as ground truth to set the image property, python version, etc.
    labels = labels or dict()
    if labels.get("system.chipType") == "GPU":
        cpu_or_gpu = "GPU"
    elif labels.get("system.chipType") == "CPU":
        cpu_or_gpu = "CPU"
    py_version = labels.get(ImageLabel.PYTHON_VERSION)

    # TODO: get the framework name from Image Label
    # extract framework name from image repo
    if repo_name.endswith("-inference"):
        framework_name = repo_name[:-10]
    elif repo_name.endswith("-training"):
        framework_name = repo_name[:-9]
    else:
        framework_name = repo_name

    framework_name = _NORMALIZED_FRAMEWORK_NAMES.get(framework_name, framework_name)
    image_name = image_obj["Name"]

    return ImageInfo(
        image_name=image_name,
        image_uri=image_uri,
        framework_name=framework_name,
        framework_version=fw_version,
        accelerator_type=cpu_or_gpu,
        python_version=py_version,
        image_scope=image_scope,
    )


def _list_images(
    labels: List[str],
    session: Session,
    name: Optional[str] = None,
    page_number=1,
    page_size=50,
):
    gen = make_list_resource_iterator(
        session.image_api.list,
        name=name,
        labels=labels,
        verbose=True,
        # set the workspace_id manually, prevent using the default workspace of the
        # session.
        workspace_id=0,
        order="DESC",
        sort_by="GmtCreateTime",
        page_number=page_number,
        page_size=page_size,
    )
    return gen


def retrieve(
    framework_name: str,
    framework_version: str,
    accelerator_type: str = "CPU",
    image_scope: Optional[str] = ImageScope.TRAINING,
    session: Optional[Session] = None,
) -> ImageInfo:
    """Get a container image URI that satisfies the specified requirements.

    Examples::

        # get a TensorFlow image with specific version for training.
        retrieve(framework_name="TensorFlow", framework_version="2.3")

        # get the latest PyTorch image that supports GPU for inference.
        retrieve(
            framework_name="PyTorch",
            framework_version="latest",
            accelerator_type="GPU",
            scope=ImageScope.INFERENCE,
        )

    Args:
        framework_name (str): The name of the framework. Possible values include
            TensorFlow, XGBoost, PyTorch, OneFlow, and others.
        framework_version (str): The version of the framework to use. Get the latest
            version supported in PAI by set the parameters as 'latest'.
        image_scope (str, optional): The scope of the image to use. Possible values
            include 'training', 'inference', and 'develop'.
        accelerator_type (str, optional): The name of the accelerator to use. Possible
            values including 'CPU', and 'GPU', (Default CPU).
        session (:class:`pai.session.Session`, optional): A session object to interact
            with the PAI Service. If not provided, a default session will be used.

    Returns:
        ImageInfo: A object contains information of the image that satisfy the
            requirements.

    Raises:
        RuntimeError: A RuntimeErrors is raised if the specific image is not found.
    """
    session = session or get_default_session()
    framework_name = framework_name.lower()
    supports_fw = [fw.lower() for fw in SUPPORTED_IMAGE_FRAMEWORKS]
    if framework_name not in supports_fw:
        raise ValueError(
            f"The framework ({framework_name}) is not supported by the"
            f" retrieve method: supported frameworks"
            f" {', '.join(SUPPORTED_IMAGE_FRAMEWORKS)}",
        )

    # label filter used to list official images of specific scope.
    labels = [ImageLabel.OFFICIAL_LABEL, ImageScope.to_image_label(image_scope)]

    # if accelerator_type is not specified, use CPU image by default.
    if not accelerator_type or accelerator_type.lower() == "cpu":
        labels.append(ImageLabel.DEVICE_TYPE_CPU)
    elif accelerator_type.lower() == "gpu":
        labels.append(ImageLabel.DEVICE_TYPE_GPU)
    else:
        raise ValueError(
            f"Given accelerator type ({accelerator_type}) is not supported, only"
            f" CPU and GPU is supported."
        )

    resp = _list_images(name=framework_name, labels=labels, session=session)

    # extract image properties, such as framework version, py_version, os_version, etc,
    # from image tag.
    candidates = []
    for image_item in resp:
        image_info = _make_image_info(
            image_obj=image_item,
            image_scope=image_scope,
        )
        if image_info.framework_name.lower() != framework_name.lower():
            continue
        candidates.append(image_info)

    if not candidates:
        raise RuntimeError(
            f"Not found any image that satisfy the requirements: framework_name="
            f"{framework_name}, accelerator={accelerator_type}"
        )

    if framework_version.lower() == "latest":
        # select the latest framework version.
        candidates = sorted(
            candidates,
            key=lambda img: to_semantic_version(img.framework_version),
            reverse=True,
        )
        return candidates[0]
    else:
        # find the image with the specific framework version.
        img = next(
            (img for img in candidates if img.framework_version == framework_version),
            None,
        )
        if not img:
            supported_versions = [img.framework_version for img in candidates]
            raise RuntimeError(
                f"Not found the specific framework: framework_name={framework_name}, "
                f"framework_version={framework_version}, supported versions for the"
                f" framework are {','.join(supported_versions)} "
            )
        else:
            return img


def list_images(
    framework_name: str,
    session: Optional[Session] = None,
    image_scope: Optional[str] = ImageScope.TRAINING,
) -> List[ImageInfo]:
    """List available images provided by PAI.

    Args:
        framework_name (str): The name of the framework. Possible values include
            TensorFlow, XGBoost, PyTorch, OneFlow, and others.
        image_scope (str, optional): The scope of the image to use. Possible values
            include 'training', 'inference', and 'develop'.
        session (:class:`pai.session.Session`): A session object used to interact with
            the PAI Service. If not provided, a default session is used.

    Returns:
        List[ImageInfo]: A list of image URIs.

    """
    session = session or get_default_session()
    if not framework_name or not framework_name.strip():
        framework_name = None
    else:
        framework_name = framework_name.strip().lower()

    labels = [
        ImageScope.to_image_label(image_scope),
        ImageLabel.OFFICIAL_LABEL,
    ]
    images = _list_images(labels=labels, session=session)

    images = [
        _make_image_info(
            item,
            image_scope=image_scope,
        )
        for item in images
    ]
    if framework_name:
        return [img for img in images if img.framework_name.lower() == framework_name]
    else:
        return images



#  Copyright 2024 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import distutils.dir_util
import json
import os
import posixpath
import re
import shlex
import shutil
import tempfile
import textwrap
import typing
from typing import Any, Dict

from pai.session import Session

from ..common.consts import INSTANCE_TYPE_LOCAL_GPU
from ..common.docker_utils import ContainerRun, run_container
from ..common.logging import get_logger
from ..common.oss_utils import OssUriObj, download, is_oss_uri

if typing.TYPE_CHECKING:
    from ..estimator import Estimator


logger = get_logger(__name__)


class _TrainingEnv(object):
    ENV_PAI_HPS = "PAI_HPS"
    ENV_PAI_HPS_PREFIX = "PAI_HPS_"
    ENV_PAI_USER_ARGS = "PAI_USER_ARGS"
    ENV_PAI_INPUT_PREFIX = "PAI_INPUT_"
    ENV_PAI_OUTPUT_PREFIX = "PAI_OUTPUT_"
    ENV_PAI_WORKING_DIR = "PAI_WORKING_DIR"


class _TrainingJobConfig(object):
    WORKING_DIR = "/ml/usercode/"
    INPUT_CONFIG_DIR = "/ml/input/config/"
    INPUT_DATA_DIR = "/ml/input/data/"
    OUTPUT_DIR = "/ml/output/"


_ENV_NOT_ALLOWED_CHARS = re.compile(r"[^a-zA-Z0-9_]")
_TRAINING_LAUNCH_SCRIPT_TEMPLATE = textwrap.dedent(
    """\
#!/bin/sh

env

# change to working directory
if [ -n "$PAI_WORKING_DIR" ]; then
    echo "Change to Working Directory", $PAI_WORKING_DIR
    mkdir -p $PAI_WORKING_DIR && cd $PAI_WORKING_DIR
fi

# install requirements
if [ -e "requirements.txt" ]; then
    echo "Installing dependencies from requirements.txt"
    python -m pip install -r requirements.txt
fi

echo "User program launching"
echo "-----------------------------------------------------------------"

sh {0}
"""
)


class LocalTrainingJob(object):
    """A class that represents a local training job running with docker container."""

    def __init__(
        self,
        estimator: "Estimator",
        inputs: Dict[str, Any],
        instance_type: str = None,
        temp_dir: str = None,
        job_name: str = None,
    ):
        self.estimator = estimator
        self.inputs = inputs
        self.tmp_dir = temp_dir or tempfile.mkdtemp()
        self.job_name = job_name
        self.instance_type = instance_type
        logger.info("Local TrainingJob temporary directory: {}".format(self.tmp_dir))
        self._container_run: ContainerRun = None

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self._container_run:
            container = self._container_run.container
            container_name, container_id, status = (
                container.name,
                container.id,
                container.status,
            )
        else:
            container_name, container_id, status = None, None, None
        return f"LocalTrainingJob(container_name={container_name}, container_id={container_id}, status={status})"

    @property
    def session(self) -> Session:
        return self.estimator.session

    def prepare_env(self) -> Dict[str, str]:
        """Prepare environment variables for the training job."""

        # Hyperparameters environment variables
        def _normalize_name(name: str) -> str:
            # replace all non-alphanumeric characters with underscore
            return _ENV_NOT_ALLOWED_CHARS.sub("_", name).upper()

        env = {}
        user_args = []
        for name, value in self.estimator.hyperparameters.items():
            env[_TrainingEnv.ENV_PAI_HPS_PREFIX + _normalize_name(name)] = str(value)
            user_args.extend(["--" + name, shlex.quote(str(value))])
        env[_TrainingEnv.ENV_PAI_USER_ARGS] = " ".join(
            [shlex.quote(v) for v in user_args]
        )
        env[_TrainingEnv.ENV_PAI_HPS] = json.dumps(
            {name: str(value) for name, value in self.estimator.hyperparameters.items()}
        )

        # Environments for input channel
        for name, value in self.inputs.items():
            if (is_oss_uri(value) and value.endswith("/")) or os.path.isdir(value):
                env[
                    _TrainingEnv.ENV_PAI_INPUT_PREFIX + _normalize_name(name)
                ] = posixpath.join(_TrainingJobConfig.INPUT_DATA_DIR, name)
            else:
                file_name = os.path.basename(value)
                env[
                    _TrainingEnv.ENV_PAI_INPUT_PREFIX + _normalize_name(name)
                ] = posixpath.join(_TrainingJobConfig.INPUT_DATA_DIR, name, file_name)

        # Environments for output channel.
        # By default, TrainingJob invoked by Estimator will have two output channels:
        # 'model' and 'checkpoints'
        output_channel = ["model", "checkpoints"]
        for name in output_channel:
            env[
                _TrainingEnv.ENV_PAI_OUTPUT_PREFIX + _normalize_name(name)
            ] = posixpath.join(_TrainingJobConfig.OUTPUT_DIR, name)

        env[_TrainingEnv.ENV_PAI_WORKING_DIR] = _TrainingJobConfig.WORKING_DIR
        return env

    def run(self):
        """Run estimator job in local with docker."""
        output_model_path = self.output_path()
        os.makedirs(output_model_path, exist_ok=True)
        volumes = {}

        tmp_dir = tempfile.mkdtemp()
        # 1. Prepare source code to directory /ml/usercode
        user_code_dir = os.path.join(self.tmp_dir, "user_code")
        if is_oss_uri(self.estimator.source_dir):
            raise RuntimeError("OSS source code is not supported in local training.")
        shutil.copytree(self.estimator.source_dir, user_code_dir)
        volumes[user_code_dir] = {
            "bind": _TrainingJobConfig.WORKING_DIR,
            "mode": "rw",
        }

        # 2. Prepare input data for training job.
        input_data = self.prepare_input_data()
        for host_path, container_path in input_data.items():
            volumes[host_path] = {
                "bind": container_path,
                "mode": "rw",
            }

        # 3. Prepare input config files, such as hyperparameters.json,
        # training-job.json, etc.
        input_config_path = os.path.join(tmp_dir, "config")
        os.makedirs(input_config_path, exist_ok=True)
        self.prepare_input_config(input_config_path=input_config_path)
        volumes[input_config_path] = {
            "bind": _TrainingJobConfig.INPUT_CONFIG_DIR,
            "mode": "rw",
        }

        execution_dir = os.path.join(tmp_dir, "config", "execution")
        os.makedirs(execution_dir, exist_ok=True)
        command_path = os.path.join(execution_dir, "command.sh")
        with open(command_path, "w") as f:
            f.write(self.estimator.command)
        launch_script_path = os.path.join(input_config_path, "launch.sh")
        with open(launch_script_path, "w") as f:
            f.write(
                _TRAINING_LAUNCH_SCRIPT_TEMPLATE.format(
                    posixpath.join(
                        _TrainingJobConfig.INPUT_CONFIG_DIR, "execution/command.sh"
                    )
                )
            )

        # 4. Config output model channel
        volumes[output_model_path] = {
            "bind": posixpath.join(_TrainingJobConfig.OUTPUT_DIR, "model"),
            "mode": "rw",
        }

        gpu_count = (
            -1 if self.instance_type.strip() == INSTANCE_TYPE_LOCAL_GPU else None
        )
        self._container_run = run_container(
            environment_variables=self.prepare_env(),
            image_uri=self.estimator.image_uri,
            entry_point=[
                "/bin/sh",
                posixpath.join(_TrainingJobConfig.INPUT_CONFIG_DIR, "launch.sh"),
            ],
            volumes=volumes,
            working_dir=_TrainingJobConfig.WORKING_DIR,
            gpu_count=gpu_count,
        )

    def prepare_input_config(self, input_config_path):
        """Prepare input config for TrainingJob, such as hyperparameters.json,
        trainingjob.json."""
        with open(os.path.join(input_config_path, "hyperparameters.json"), "w") as f:
            hps = self.estimator.hyperparameters or dict()
            f.write(json.dumps({k: str(v) for k, v in hps.items()}))

    def prepare_input_data(self) -> Dict[str, str]:
        """Prepare input data config."""
        input_data_configs = {}

        for name, input_data in self.inputs.items():
            local_channel_path = os.path.join(self.tmp_dir, f"input/data/{name}")
            os.makedirs(local_channel_path, exist_ok=True)
            input_data_configs[local_channel_path] = posixpath.join(
                _TrainingJobConfig.INPUT_DATA_DIR, name
            )
            if is_oss_uri(input_data):
                oss_uri_obj = OssUriObj(input_data)
                oss_bucket = self.session.get_oss_bucket(oss_uri_obj.bucket_name)
                os.makedirs(local_channel_path, exist_ok=True)
                download(
                    oss_uri_obj.object_key,
                    local_path=local_channel_path,
                    bucket=oss_bucket,
                )
                input_data_configs[local_channel_path] = posixpath.join(
                    _TrainingJobConfig.INPUT_DATA_DIR, name
                )
            else:
                # If the input data is local files, copy the input data to a
                # temporary directory.
                if not os.path.exists(input_data):
                    raise ValueError(
                        "Input data not exists: name={} input_data={}".format(
                            name, input_data
                        )
                    )
                elif os.path.isdir(input_data):
                    distutils.dir_util.copy_tree(input_data, local_channel_path)
                else:
                    shutil.copy(
                        input_data,
                        os.path.join(local_channel_path, os.path.basename(input_data)),
                    )

        return input_data_configs

    def wait(self, show_logs: bool = True):
        self._container_run.watch(show_logs=show_logs)

    def output_path(self, channel_name="model"):
        return os.path.join(self.tmp_dir, "output", f"{channel_name}/")

    def is_succeeded(self):
        """Return True if the training job is succeeded, otherwise return False."""
        return self._container_run.is_succeeded()



#  Copyright 2023 Alibaba, Inc. or its affiliates.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional, Union

from alibabacloud_credentials.client import Client as CredentialClient
from alibabacloud_sts20150401.client import Client as StsClient

from ..common.consts import DEFAULT_NETWORK_TYPE, PAI_VPC_ENDPOINT, Network
from ..common.utils import is_domain_connectable
from ..libs.alibabacloud_pai_dsw20220101.client import Client as DswClient
from .algorithm import AlgorithmAPI
from .base import PAIRestResourceTypes, ServiceName, WorkspaceScopedResourceAPI
from .client_factory import ClientFactory
from .code_source import CodeSourceAPI
from .dataset import DatasetAPI
from .experiment import ExperimentAPI
from .image import ImageAPI
from .job import JobAPI
from .model import ModelAPI
from .pipeline import PipelineAPI
from .pipeline_run import PipelineRunAPI
from .service import ServiceAPI
from .tensorboard import TensorBoardAPI
from .training_job import TrainingJobAPI
from .workspace import WorkspaceAPI

_RESOURCE_API_MAPPING = {
    PAIRestResourceTypes.DlcJob: JobAPI,
    PAIRestResourceTypes.CodeSource: CodeSourceAPI,
    PAIRestResourceTypes.Dataset: DatasetAPI,
    PAIRestResourceTypes.Image: ImageAPI,
    PAIRestResourceTypes.Service: ServiceAPI,
    PAIRestResourceTypes.Model: ModelAPI,
    PAIRestResourceTypes.Workspace: WorkspaceAPI,
    PAIRestResourceTypes.Algorithm: AlgorithmAPI,
    PAIRestResourceTypes.TrainingJob: TrainingJobAPI,
    PAIRestResourceTypes.Pipeline: PipelineAPI,
    PAIRestResourceTypes.PipelineRun: PipelineRunAPI,
    PAIRestResourceTypes.TensorBoard: TensorBoardAPI,
    PAIRestResourceTypes.Experiment: ExperimentAPI,
}


class ResourceAPIsContainerMixin(object):
    """ResourceAPIsContainerMixin provides Resource Operation APIs."""

    _credential_client = None
    _credential_config = None
    _region_id = None
    _workspace_id = None

    def __init__(
        self, header=None, runtime=None, network: Optional[Union[str, Network]] = None
    ):
        """Initialize ResourceAPIsContainerMixin.

        Args:
            header: Header for API request.
            runtime: Runtime for API request.
            network: Network type used to connect to PAI services.
        """
        self.header = header
        self.runtime = runtime
        self.api_container = dict()
        self.acs_client_container = dict()
        if network:
            self.network = (
                Network.from_string(network) if isinstance(network, str) else network
            )
        elif DEFAULT_NETWORK_TYPE:
            self.network = Network.from_string(DEFAULT_NETWORK_TYPE)
        else:
            self.network = (
                Network.VPC
                if is_domain_connectable(PAI_VPC_ENDPOINT.format(self._region_id))
                else Network.PUBLIC
            )

    def _acs_credential_client(self):
        if self._credential_client:
            return self._credential_client
        self._credential_client = CredentialClient(config=self._credential_config)
        return self._credential_client

    def _get_acs_client(self, service_name):
        if service_name in self.acs_client_container:
            return self.acs_client_container[service_name]
        acs_client = ClientFactory.create_client(
            service_name=service_name,
            credential_client=self._acs_credential_client(),
            region_id=self._region_id,
            network=self.network,
        )
        self.acs_client_container[service_name] = acs_client
        return acs_client

    @property
    def _acs_workspace_client(self):
        return self._get_acs_client(ServiceName.PAI_WORKSPACE)

    @property
    def _acs_dlc_client(self):
        return self._get_acs_client(ServiceName.PAI_DLC)

    @property
    def _acs_paiflow_client(self):
        return self._get_acs_client(ServiceName.PAIFLOW)

    @property
    def _acs_eas_client(self):
        return self._get_acs_client(ServiceName.PAI_EAS)

    @property
    def _acs_training_client(self):
        return self._get_acs_client(ServiceName.PAI_STUDIO)

    @property
    def _acs_sts_client(self) -> StsClient:
        return self._get_acs_client(ServiceName.STS)

    @property
    def _acs_dsw_client(self) -> DswClient:
        return self._get_acs_client(ServiceName.PAI_DSW)

    def get_api_by_resource(self, resource_type):
        if resource_type in self.api_container:
            return self.api_container[resource_type]

        api_cls = _RESOURCE_API_MAPPING[resource_type]
        acs_client = self._get_acs_client(api_cls.BACKEND_SERVICE_NAME)
        if issubclass(api_cls, WorkspaceScopedResourceAPI):
            api = api_cls(
                workspace_id=self._workspace_id,
                acs_client=acs_client,
                header=self.header,
                runtime=self.runtime,
            )
        elif api_cls == ServiceAPI:
            # for PAI-EAS service api, we need to pass region_id.
            api = api_cls(
                acs_client=acs_client,
                region_id=self._region_id,
                header=self.header,
                runtime=self.runtime,
            )
        else:
            api = api_cls(
                acs_client=acs_client,
                header=self.header,
                runtime=self.runtime,
            )
        self.api_container[resource_type] = api
        return api

    @property
    def job_api(self) -> JobAPI:
        """Returns JobAPI for job operation."""
        return self.get_api_by_resource(PAIRestResourceTypes.DlcJob)

    @property
    def tensorboard_api(self) -> TensorBoardAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.TensorBoard)

    @property
    def code_source_api(self) -> CodeSourceAPI:
        """Return CodeSource API for code_source operation"""
        return self.get_api_by_resource(PAIRestResourceTypes.CodeSource)

    @property
    def dataset_api(self) -> DatasetAPI:
        """Return Dataset API for dataset operation"""
        return self.get_api_by_resource(PAIRestResourceTypes.Dataset)

    @property
    def image_api(self) -> ImageAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Image)

    @property
    def model_api(self) -> ModelAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Model)

    @property
    def service_api(self) -> ServiceAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Service)

    @property
    def workspace_api(self) -> WorkspaceAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Workspace)

    @property
    def algorithm_api(self) -> AlgorithmAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Algorithm)

    @property
    def training_job_api(self) -> TrainingJobAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.TrainingJob)

    @property
    def pipeline_api(self) -> PipelineAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Pipeline)

    @property
    def pipeline_run_api(self) -> PipelineRunAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.PipelineRun)

    @property
    def experiment_api(self) -> ExperimentAPI:
        return self.get_api_by_resource(PAIRestResourceTypes.Experiment)
